#!/bin/bash

prefix="/opt/cyops-integrations/"
pg_version="12"
pg_hba_file="/var/lib/pgsql/$pg_version/data/pg_hba.conf"
product_version="@product_version@"
pg_password="@pg_password@"
product_yum_server="update.cybersponse.com"
secure_message_exchange_host="@secure_message_exchange_hostname@"

# Check root
if [[ $EUID -ne 0 ]]; then
   echo "This script must be run as root. Please sudo su and re-run script"
   exit 1
fi

if ! curl -I --silent --connect-timeout 10 https://$product_yum_server -k > /dev/null; then
   echo "======================================================================================================";
   echo "Please ensure connectivity to $product_yum_server for agent installation"
   echo "======================================================================================================";
   exit 1;
fi

if ! getent hosts $secure_message_exchange_host; then
   echo "==================================================================================================";
   echo "The Secure Message Exchange server $rabbitmq_host is not resolvable from this host."
   echo "It is recommended to establish the connectivity to the router before running the agent installer."
   echo "==================================================================================================";
   read -p "Exit the installer (y/n)?" -n 1 -r
   if [[ $REPLY =~ ^[Yy]$ ]]
   then
       echo '\n'
       exit 1;
   fi
fi

app_repo="[cs-app]
name=CyberSponse Application Repository
baseurl=https://$product_yum_server/$product_version/cybersponse/x86_64/
gpgcheck=0
enabled=1"

postgres_repo="[cs-pgdg12]
name=CyberSponse PostgreSQL 12
baseurl=https://$product_yum_server/$product_version/third_party/postgres/12/redhat/rhel-7-x86_64/
enabled=1
gpgcheck=1
gpgkey=https://$product_yum_server/$cs_version/third_party/postgres/RPM-GPG-KEY-PGDG-12"

echo "$app_repo" > /etc/yum.repos.d/cs-app.repo
echo "$postgres_repo" > /etc/yum.repos.d/cs-postgres.repo

sudo chmod 644 /etc/yum.repos.d/cs-app.repo
sudo chmod 644 /etc/yum.repos.d/cs-postgres.repo

yum clean all

rpm -q postgresql12-server > /dev/null
if [ $? -ne 0 ]; then
    echo "======================================================================================================";
    echo "Installing Postgres Server"
    echo "======================================================================================================";
    set -e
    yum install -y postgresql12-server --nogpgcheck
    set +e

    #pg setup
    /usr/pgsql-12/bin/postgresql-12-setup initdb
    systemctl enable postgresql-12
    pg_hba='@pg_hba@'
    # Changing the postgres access method to password
    > /var/lib/pgsql/12/data/pg_hba.conf
    echo "$pg_hba" > /var/lib/pgsql/12/data/pg_hba.conf
    chown postgres:postgres /var/lib/pgsql/12/data/pg_hba.conf
    chmod 600 /var/lib/pgsql/12/data/pg_hba.conf

    systemctl restart postgresql-12.service

    # create pg user
    sudo -Hiu postgres psql -c 'DROP USER IF EXISTS cyberpgsql;'
    sudo -Hiu postgres psql -U postgres -c "CREATE USER cyberpgsql WITH SUPERUSER PASSWORD '$pg_password'"
    sudo -Hiu postgres psql -c 'DROP DATABASE IF EXISTS connectors;'
    sudo -Hiu postgres createdb connectors 'cyberpgsql'
fi

rpm -q cyops-integrations-agent > /dev/null
if [ $? -ne 0 ]; then
    set -e
    yum install -y cyops-integrations-agent --nogpgcheck
    set +e
    systemctl enable cyops-integrations-agent
fi

rm -f /etc/yum.repos.d/cs-*

agent_config='@agent_config@'
ca_cert='@ca_cert@'
connectors='@connectors@'

echo "$agent_config" > /opt/cyops-integrations/integrations/configs/agent_config.yml
echo "$ca_cert" > /opt/cyops-integrations/integrations/configs/ca_cert.pem
echo "$connectors" > /opt/cyops-integrations/integrations/connectors.json

# restart service
systemctl restart postgresql-12
systemctl restart cyops-integrations-agent

# run migration
cd $prefix/integrations
$prefix/.env/bin/python3 manage.py migrate


rpm -q gcc > /dev/null

if [ $? -ne 0 ]; then
   echo "======================================================================================================";
   echo "attempting to install gcc"
   echo "======================================================================================================";
   yum install -y gcc > /dev/null 2>&1

   if [ 0 -ne  $? ]; then
      echo "======================================================================================================";
      echo "Failed to install GCC rpm. Some of the FortiSOAR connectors may have a dependency on gcc."
      echo "It is recommended to install gcc by enabling the CentOS yum repositories on this instance and running 'yum install gcc -y'"
      echo "======================================================================================================";
   else
      echo "successfully installed gcc"
   fi
fi

/opt/cyops-integrations/.env/bin/python /opt/cyops-integrations/integrations/manage.py install_connector > /dev/null

echo "Agent Deployed Successfully"
