import uuid
import json
import logging
import requests
import time
from django.conf import settings
from utils.config_parser import all_config
from urllib.parse import urlparse

logger = logging.getLogger('connectors')


def make_request(url, method, body=None, __async=False, *args, **kwargs):
    from postman.core.publisher import Publisher
    if not settings.MASTER_ID: _populated_Related_info()

    if not bool(urlparse(url).netloc):
        url = settings.HOST_URL + str(url)
    else:
        url = url
    if not body and __async: body = {'__async': __async}
    data = {
        'url': url,
        'body': body,
        'method': method,
        '__async': __async
    }
    data.update(**kwargs)


    pub_config = {
        'exchangeName': 'dexchange.cyops.postman.{0}'.format(settings.MASTER_ID),
        'routingKey': 'route.postman.chapi.remoterequest.{0}'.format(settings.MASTER_ID)
    }
    message_json = {
        "data": data,
        "sourceId": settings.SELF_ID,
        "destinationId": settings.MASTER_ID,
        "method": method
    }
    callback_queue = 'route.postman.chapi.remoteresponse.{0}'.format(settings.SELF_ID)
    correlation_id = str(uuid.uuid4())
    publisher = Publisher(pub_config)
    publisher.publish(json.dumps(message_json), settings.MASTER_ID, correlation_id=correlation_id,
                      callback_queue=callback_queue)

    if __async: return {'task_id': 'dummy'}

    timeout = time.time() + 60
    while not settings.CH_API_REQUEST.get(correlation_id) and timeout > time.time():
        pass

    response = settings.CH_API_REQUEST.pop(correlation_id, {})

    if response.get('status') == 'success':
        return response.get('result')
    else:
        msg = response.get('message', 'Request timeout')
        logger.error(msg)
        raise requests.exceptions.HTTPError(msg)


def _populated_Related_info():
    settings.MASTER_ID = all_config.get('cyops.instance.masterId')
    settings.SELF_ID = all_config.get('cyops.instance.agentId')


def maybe_json_or_raise(response):
    """
    Helper function for processing request responses

    Returns any json found in the response. Otherwise, it will extract the
    response as text, or, failing that, as bytes.

    :return: the response from the request
    :rtype: dict or str or bytes
    :raises: :class:`requests.HTTPError` if status code was 4xx or 5xx
    """
    if response.ok:
        try:
            logger.info('Processing request responses.')
            return response.json(strict=False)
        except Exception:
            return response.text or response.content
    else:
        msg = ''
        try:
            msg = response.json()
            logger.warn(msg)
        except Exception:
            pass
        if not msg:
            msg = response.text
            logger.warn(msg)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # add any response content to the error message
            msg = '{} :: {} :: Url: {}'.format(str(e), msg, response.url)
            logger.error(msg)
            raise requests.exceptions.HTTPError(msg, response=response)


def make_file_upload_request(*args, **kwargs):
    return None
